@@ -1,10 +1,14 @@
-from typing import Iterable, List, Mapping, Optional, Union
+from __future__ import annotations
+
+from typing import TYPE_CHECKING, Iterable, List, Mapping, Optional, Union
 
-import polars as pl
 import pyarrow as pa
 
-from chalk.client.client_impl import ChalkConfigurationException
 from chalk.features.dataframe import DataFrame
+from chalk.utils.missing_dependency import missing_dependency_exception
+
+if TYPE_CHECKING:
+    import polars as pl
 
 try:
     import duckdb
@@ -32,7 +36,7 @@ def _resolve_implicit_col_feature_mappings_str(
 
 def query_arrow_table(query: str, table: pa.Table) -> pa.Table:
     if duckdb is None or sqlglot is None:
-        raise ChalkConfigurationException.missing_dependency("chalkpy[sql]")
+        raise missing_dependency_exception("chalkpy[sql]")
     t: sqlglot.exp.Table
     table_names = list({t.name for t in sqlglot.parse_one(query).find_all(sqlglot.exp.Table)})
     con = duckdb.connect(database=":memory:")
@@ -49,6 +53,10 @@ def query_as_features(
     fqn_to_name: Mapping[str, str],
     table: pa.Table,
 ) -> pl.DataFrame:
+    try:
+        import polars as pl
+    except ImportError:
+        raise missing_dependency_exception("chalkpy[polars]")
     res = query_arrow_table(query, table)
     col_feature_mappings = _resolve_implicit_col_feature_mappings_str(
         # Note: snowflake by default capitalizes all table and column names, although it's
@@ -73,6 +81,11 @@ def query_as_feature_formatted(
     table: Union[pa.Table, DataFrame, pl.DataFrame, pl.LazyFrame],
 ) -> pl.DataFrame:
     # We're going to string replace whatever was there before
+    try:
+        import polars as pl
+    except ImportError:
+        raise missing_dependency_exception("chalkpy[polars]")
+
     unformatted = formatted_query.replace(str(table), "__chalk_table__")
     pa_table = table
     if isinstance(table, DataFrame):
