@@ -1,19 +1,17 @@
 from __future__ import annotations
 
-import enum
-import json
+import collections.abc
 import time
-from concurrent.futures import Future
-from datetime import datetime
-from itertools import chain
-from typing import Any, Iterable, List, Mapping, Optional, Sequence, Type, TypeVar, Union
+from datetime import datetime, timezone
+from typing import TYPE_CHECKING, Any, Iterable, List, Mapping, Optional, Sequence, Type, TypeVar, Union
 from urllib.parse import urljoin
 
-import polars as pl
+import pandas as pd
 import requests
 from pydantic import BaseModel, ValidationError
 from requests import HTTPError
 
+from chalk._version import __version__ as chalkpy_version
 from chalk.client.client_protocol import (
     ChalkAPIClientProtocol,
     ChalkBaseException,
@@ -27,15 +25,17 @@ from chalk.client.client_protocol import (
 )
 from chalk.client.dataset import DatasetVersion, load_dataset
 from chalk.config.auth_config import load_token
-from chalk.features import DataFrame, Feature, Features, FeatureWrapper, ensure_feature, unwrap_feature
+from chalk.features import DataFrame, Feature, FeatureWrapper, ensure_feature
 from chalk.features.pseudofeatures import CHALK_TS_FEATURE
 from chalk.serialization.codec import FEATURE_CODEC
 from chalk.utils.log_with_context import get_logger
+from chalk.utils.missing_dependency import missing_dependency_exception
+
+if TYPE_CHECKING:
+    import polars as pl
 
 _logger = get_logger(__name__)
 
-import pandas as pd
-
 
 class _ExchangeCredentialsRequest(BaseModel):
     client_id: str
@@ -47,6 +47,9 @@ class _ExchangeCredentialsResponse(BaseModel):
     access_token: str
     token_type: str
     expires_in: int
+    # expires_at: datetime
+    api_server: str
+    primary_environment: Optional[str]
     engines: Optional[Mapping[str, str]] = None
 
 
@@ -101,6 +104,7 @@ class _CreateOfflineQueryJobRequest(BaseModel):
     """
 
     output: List[str]
+    required_output: List[str]
     destination_format: str
     input: Optional[_OfflineQueryInput] = None
     max_samples: Optional[int] = None
@@ -111,6 +115,27 @@ class _CreateOfflineQueryJobRequest(BaseModel):
     environment: Optional[str] = None
 
 
+# FIXME document
+class _ComputeResolverOutputRequest(BaseModel):
+    input: _OfflineQueryInput
+    resolver_fqn: str
+    branch: Optional[str] = None
+    environment: Optional[str] = None
+
+
+class _RecomputeResolverOutputRequest(BaseModel):
+    persistent_id: str
+    resolver_fqn: str
+    branch: Optional[str] = None
+    environment: Optional[str] = None
+
+
+class _ComputeResolverOutputResponse(BaseModel):
+    job_id: str
+    persistent_id: str
+    errors: Optional[List[ChalkError]]
+
+
 class _CreateOfflineQueryJobResponse(BaseModel):
     """
     Attributes:
@@ -132,17 +157,7 @@ class _GetOfflineQueryJobResponse(BaseModel):
     is_finished: bool
     version: int
     urls: List[str]
-
-
-class ChalkConfigurationException(ChalkBaseException):
-    message: str
-
-    def __init__(self, message: str):
-        super().__init__(message)
-
-    @classmethod
-    def missing_dependency(cls, name: str):
-        return cls(f"Missing pip dependency '{name}'. Please add this to your requirements.txt file and pip install.")
+    errors: Optional[List[ChalkError]]
 
 
 class ChalkOfflineQueryException(ChalkBaseException):
@@ -173,23 +188,35 @@ class ChalkDatasetDownloadException(ChalkBaseException):
 
 class OnlineQueryResponseWrapper(OnlineQueryResponse):
     data: List[FeatureResult]
-    errors: Optional[List[ChalkError]]
+    errors: List[ChalkError]
+    warnings: List[str]
 
-    def __init__(
-        self,
-        data: List[FeatureResult],
-        errors: Optional[List[ChalkError]],
-    ):
+    def __init__(self, data: List[FeatureResult], errors: List[ChalkError], warnings: List[str]):
         self.data = data
         self.errors = errors
+        self.warnings = warnings
         for d in self.data:
             if d.value is not None:
-                d.value = FEATURE_CODEC.decode_fqn(d.field, d.value)
+                try:
+                    d.value = FEATURE_CODEC.decode_fqn(d.field, d.value)
+                except Exception:
+                    self.warnings.append(
+                        f"Return data {d.field}:{d.value} cannot be decoded. Attempting to JSON decode"
+                    )
+
         self._values = {d.field: d for d in self.data}
 
+    def __repr__(self) -> str:
+        info = self.errors if self.errors else self.data
+        return repr(pd.DataFrame([vars(x) for x in info]))
+
+    def _repr_html_(self) -> str | None:
+        info = self.errors if self.errors else self.data
+        return pd.DataFrame([vars(x) for x in info])._repr_html_()
+
     def get_feature(self, feature: Any) -> Optional[FeatureResult]:
         # Typing `feature` as Any, as the Features will be typed as the underlying datatypes, not as Feature
-        return self._values.get(ensure_feature(feature).root_fqn)
+        return self._values.get(str(feature))
 
     def get_feature_value(self, feature: Any) -> Optional[Any]:
         # Typing `feature` as Any, as the Features will be typed as the underlying datatypes, not as Feature
@@ -197,30 +224,6 @@ class OnlineQueryResponseWrapper(OnlineQueryResponse):
         return v and v.value
 
 
-def _expand_scalar_features_shallow(o: Union[Feature, FeatureWrapper, Features, str, Any]) -> Iterable[Feature]:
-    if isinstance(o, str):
-        yield Feature.from_root_fqn(o)
-        return
-    if isinstance(o, type) and issubclass(o, Features):
-        for f in o.features:
-            if f.is_scalar:
-                yield f
-        return
-    if isinstance(o, FeatureWrapper):
-        o = unwrap_feature(o)
-    if isinstance(o, Feature):
-        if o.is_scalar:
-            yield o
-            return
-        if o.is_has_one:
-            assert o.joined_class is not None
-            for f in o.joined_class.features:
-                if isinstance(f, Feature) and f.is_scalar:
-                    yield Feature.from_root_fqn(f"{o.root_fqn}.{f.name}")
-            return
-    raise ValueError(f"Unsupported feature {o}")
-
-
 class ChalkAPIClientImpl(ChalkAPIClientProtocol):
     def __init__(
         self,
@@ -259,9 +262,11 @@ class ChalkAPIClientImpl(ChalkAPIClientProtocol):
         self._default_headers = {
             "Accept": "application/json",
             "Content-Type": "application/json",
+            "User-Agent": f"chalkpy-{chalkpy_version}",
+            "X-Chalk-Client-Id": self._config.client_id,
         }
-        self._engines: Optional[Mapping[str, str]] = None
         self._exchanged_credentials = False
+        self._primary_environment = None
 
     def _exchange_credentials(self):
         _logger.debug("Performing a credentials exchange")
@@ -285,11 +290,11 @@ class ChalkAPIClientImpl(ChalkAPIClientProtocol):
         except ValidationError:
             raise HTTPError(response=resp)
         self._default_headers["Authorization"] = f"Bearer {creds.access_token}"
-        self._engines = creds.engines
+        self._primary_environment = creds.primary_environment
         self._exchanged_credentials = True
 
     def _get_headers(self, environment_override: Optional[str]):
-        x_chalk_env_id = environment_override or self._config.active_environment
+        x_chalk_env_id = environment_override or self._config.active_environment or self._primary_environment
         headers = dict(self._default_headers)  # shallow copy
         if x_chalk_env_id is not None:
             headers["X-Chalk-Env-Id"] = x_chalk_env_id
@@ -310,13 +315,7 @@ class ChalkAPIClientImpl(ChalkAPIClientProtocol):
             exchanged_credentials = True
             self._exchange_credentials()
         headers = self._get_headers(environment_override=environment_override)
-        active_env = environment_override or self._config.active_environment
-        if use_engine and self._engines is not None and active_env in self._engines:
-            assert isinstance(active_env, str)
-            base = self._engines[active_env]
-        else:
-            base = self._config.api_server
-        url = urljoin(base, uri)
+        url = urljoin(self._config.api_server, uri)
         json_body = json and json.dict()
         r = requests.request(method=method, headers=headers, url=url, json=json_body)
         if r.status_code in (401, 403) and not exchanged_credentials:
@@ -333,7 +332,7 @@ class ChalkAPIClientImpl(ChalkAPIClientProtocol):
 
     def upload_features(
         self,
-        input: Mapping[Union[str, Feature, Any], Any],
+        input: Mapping[Union[str, Feature, FeatureWrapper, Any], Any],
         context: Optional[OnlineQueryContext] = None,
         preview_deployment_id: Optional[str] = None,
         correlation_id: Optional[str] = None,
@@ -362,13 +361,17 @@ class ChalkAPIClientImpl(ChalkAPIClientProtocol):
         query_name: Optional[str] = None,
         meta: Optional[Mapping[str, str]] = None,
     ) -> OnlineQueryResponse:
+        warnings: List[str] = []
         encoded_inputs = {}
         for feature, value in input.items():
-            feature = ensure_feature(feature)
-            encoded_inputs[feature.fqn] = FEATURE_CODEC.encode(feature, value)
-        outputs: List[str] = []
-        for x in output:
-            outputs.extend(f.root_fqn for f in _expand_scalar_features_shallow(x))
+            try:
+                feature = ensure_feature(feature)
+                encoded_inputs[feature.fqn] = FEATURE_CODEC.encode(feature, value)
+            except Exception:
+                warnings.append(f"Input {feature} not recognized. JSON encoding {feature} and requesting anyways")
+                encoded_inputs[feature] = value
+        outputs: List[str] = [str(feature) for feature in output]
+
         request = _QueryRequest(
             inputs=encoded_inputs,
             outputs=outputs,
@@ -388,110 +391,135 @@ class ChalkAPIClientImpl(ChalkAPIClientProtocol):
             use_engine=preview_deployment_id is None,
             environment_override=context.environment if context else None,
         )
-        return OnlineQueryResponseWrapper(
-            data=resp.data,
-            errors=resp.errors,
-        )
+        return OnlineQueryResponseWrapper(data=resp.data, errors=resp.errors or [], warnings=warnings)
 
     def get_training_dataframe(
         self,
-        input: Union[Mapping[Union[str, Feature], Any], pl.DataFrame, pd.DataFrame, DataFrame],
-        input_times: List[datetime],
-        output: List[Union[str, Feature, Any]],
+        input: Union[Mapping[Union[str, Feature, FeatureWrapper, Any], Any], pd.DataFrame, pl.DataFrame, DataFrame],
+        input_times: Union[Sequence[datetime], datetime, None] = None,
+        output: Sequence[Union[str, Feature, FeatureWrapper, Any]] = (),
+        required_output: Sequence[Union[str, Feature, FeatureWrapper, Any]] = (),
+        output_ts: bool = False,
         context: Optional[OfflineQueryContext] = None,
         dataset: Optional[str] = None,
         branch: Optional[str] = None,
+        max_samples: Optional[int] = None,
     ) -> pd.DataFrame:
-        if not isinstance(input, DataFrame):
-            input = DataFrame(input)
+        try:
+            import polars as pl
+        except ImportError:
+            raise missing_dependency_exception("chalkpy[polars]")
+        if isinstance(input, (DataFrame, pl.DataFrame)):
+            input = input.to_pandas()
+
+        if isinstance(input, collections.abc.Mapping):
+            input = {str(k): v for (k, v) in input.items()}
+
+        if not isinstance(input, pd.DataFrame):
+            input = pd.DataFrame(input)
+
+        if len(output) == 0 and len(required_output) == 0:
+            raise ValueError("Either 'output' or 'required_output' must be specified.")
 
         self._get_headers(environment_override=None)
 
         return self._get_training_dataframe(
-            input=input.to_pandas(),
+            input=input,
             input_times=input_times,
-            output=output,
+            optional_output=[str(x) for x in output],
+            required_output=[str(x) for x in required_output],
+            output_id=False,
+            output_ts=output_ts,
             context=context,
             dataset=dataset,
             branch=branch,
+            max_samples=max_samples,
         )
 
-    def _decode_offline_response(self, offline_query_response: _OfflineQueryResponse) -> pd.DataFrame:
-        data = {}
-        for col_index, column in enumerate(offline_query_response.output):
-            series_values = []
-            for value in column:
-                value = FEATURE_CODEC.decode_fqn(
-                    fqn=offline_query_response.columns[col_index],
-                    value=value,
-                )
-                if isinstance(value, enum.Enum):
-                    value = value.value
-                series_values.append(value)
-
-            data[offline_query_response.columns[col_index]] = pd.Series(
-                data=series_values,
-                dtype=FEATURE_CODEC.get_pandas_dtype(offline_query_response.columns[col_index]),
-            )
-        return pd.DataFrame(data)
-
     def _get_training_dataframe(
         self,
         input: pd.DataFrame,
-        input_times: List[datetime],
-        output: List[Union[str, Feature, Any]],
-        context: Optional[OfflineQueryContext] = None,
-        dataset: Optional[str] = None,
-        branch: Optional[str] = None,
+        input_times: Union[Sequence[datetime], datetime, None],
+        optional_output: List[str],
+        required_output: List[str],
+        output_id: bool,
+        output_ts: bool,
+        max_samples: Optional[int],
+        context: Optional[OfflineQueryContext],
+        dataset: Optional[str],
+        branch: Optional[str],
     ) -> pd.DataFrame:
         columns = input.columns
-        matrix = input.T.values.tolist()
+        matrix: List[List[Any]] = input.T.values.tolist()
 
-        columns_fqn = [ensure_feature(c).root_fqn for c in chain(columns, (CHALK_TS_FEATURE,))]
+        columns_fqn = [str(c) for c in (*columns, CHALK_TS_FEATURE)]
+
+        if input_times is None:
+            input_times = datetime.now(timezone.utc)
+        if isinstance(input_times, datetime):
+            input_times = [input_times for _ in range(len(input))]
 
         matrix.append([a for a in input_times])
 
         for col_index, column in enumerate(matrix):
             for row_index, value in enumerate(column):
-                matrix[col_index][row_index] = FEATURE_CODEC.encode_fqn(
-                    fqn=columns_fqn[col_index],
-                    value=value,
-                )
+                try:
+                    encoded_feature = FEATURE_CODEC.encode_fqn(
+                        fqn=columns_fqn[col_index],
+                        value=value,
+                    )
+                except (KeyError, ValueError, TypeError):
+                    encoded_feature = value
+
+                matrix[col_index][row_index] = encoded_feature
 
         query_input = _OfflineQueryInput(
             columns=columns_fqn,
             values=matrix,
         )
 
-        query_output: List[Feature] = []
-        for x in output:
-            query_output.extend(_expand_scalar_features_shallow(x))
+        response = self._create_and_await_offline_query_job(
+            optional_output=optional_output,
+            required_output=required_output,
+            query_input=query_input,
+            output_id=output_id,
+            output_ts=output_ts,
+            dataset_name=dataset,
+            branch=branch,
+            max_samples=max_samples,
+            environment=context and context.environment,
+        )
+        return response.to_pandas()
+
+    def sample(
+        self,
+        output: Sequence[Union[str, Feature, FeatureWrapper, Any]],
+        required_output: Sequence[Union[str, Feature, FeatureWrapper, Any]] = (),
+        output_id: bool = False,
+        output_ts: bool = False,
+        max_samples: Optional[int] = None,
+        dataset: Optional[str] = None,
+        branch: Optional[str] = None,
+        context: Optional[OfflineQueryContext] = None,
+    ) -> pd.DataFrame:
+        optional_output_root_fqns = [str(f) for f in output]
+        required_output_root_fqns = [str(f) for f in required_output]
+
+        if len(output) == 0 and len(required_output) == 0:
+            raise ValueError("Either 'output' or 'required_output' must be specified.")
 
         try:
             response = self._create_and_await_offline_query_job(
-                output=query_output,
-                input=query_input,
+                query_input=None,
+                optional_output=optional_output_root_fqns,
+                required_output=required_output_root_fqns,
+                max_samples=max_samples,
+                output_id=output_id,
+                output_ts=output_ts,
                 dataset_name=dataset,
                 branch=branch,
                 environment=context and context.environment,
             )
-            return response.to_pandas()
-        except HTTPError as e:
-            _logger.debug("Got HTTP Exception while processing sample query", exc_info=e)
-            raise ChalkOfflineQueryException(message=f"HTTP error while processing sample query.", errors=[])
-
-    def sample(
-        self,
-        output: List[Union[str, Feature, FeatureWrapper, Any]],
-        max_samples: Optional[int] = None,
-        context: Optional[OfflineQueryContext] = None,
-    ) -> pd.DataFrame:
-        query_output: list[Feature] = []
-        for x in output:
-            query_output.extend(_expand_scalar_features_shallow(x))
-
-        try:
-            response = self._create_and_await_offline_query_job(output=query_output, max_samples=max_samples)
         except HTTPError as e:
             _logger.debug("Got HTTP Exception while processing sample query", exc_info=e)
             raise ChalkOfflineQueryException(message=f"HTTP error while processing sample query.", errors=[])
@@ -514,13 +542,12 @@ class ChalkAPIClientImpl(ChalkAPIClientProtocol):
                 response=ResolverRunResponse,
             )
         except HTTPError as e:
-            message = str(e)
 
             detail = e.response.json().get("detail")
             if detail is not None:
-                message = detail
-
-            raise ChalkResolverRunException(message=message)
+                raise ChalkResolverRunException(message=detail)
+            else:
+                raise
 
         return response
 
@@ -532,30 +559,32 @@ class ChalkAPIClientImpl(ChalkAPIClientProtocol):
                 response=ResolverRunResponse,
             )
         except HTTPError as e:
-            message = str(e)
 
             detail = e.response.json().get("detail")
             if detail is not None:
-                message = detail
-
-            raise ChalkResolverRunException(message=message)
+                raise ChalkResolverRunException(message=detail)
+            else:
+                raise
 
         return response
 
     def _create_and_await_offline_query_job(
         self,
-        output: Sequence[Feature],
-        input: Optional[_OfflineQueryInput] = None,
-        max_samples: Optional[int] = None,
-        dataset_name: Optional[str] = None,
-        branch: Optional[str] = None,
-        environment: Optional[str] = None,
+        optional_output: List[str],
+        required_output: List[str],
+        query_input: Optional[_OfflineQueryInput],
+        max_samples: Optional[int],
+        dataset_name: Optional[str],
+        branch: Optional[str],
+        environment: Optional[str],
+        output_id: bool,
+        output_ts: bool,
     ) -> pl.DataFrame:
-        output = list(output)
         req = _CreateOfflineQueryJobRequest(
-            output=[x.root_fqn for x in output],
+            output=optional_output,
+            required_output=required_output,
             destination_format="PARQUET",
-            input=input,
+            input=query_input,
             max_samples=max_samples,
             dataset_name=dataset_name,
             branch=branch,
@@ -573,45 +602,60 @@ class ChalkAPIClientImpl(ChalkAPIClientProtocol):
             time.sleep(0.5)
         return load_dataset(
             uris=status.urls,
-            version=DatasetVersion.BIGQUERY_JOB_WITH_B32_ENCODED_COLNAMES,
-            features=output,
+            output_features=[*optional_output, *required_output],
+            version=DatasetVersion(response.version),
+            output_id=output_id,
+            output_ts=output_ts,
         )
 
-        df_futures: list[Future[pl.DataFrame]] = []
-        for f in status.urls:
-            df_futures.append(executor.submit(pl.read_parquet, f))
+    def compute_resolver_output(
+        self,
+        input: Union[Mapping[Union[str, Feature], Any], pl.DataFrame, pd.DataFrame, DataFrame],
+        input_times: List[datetime],
+        resolver: str,
+    ) -> pl.DataFrame:
+        if not isinstance(input, DataFrame):
+            input = DataFrame(input)
+        input = input.to_pandas()
 
-        dfs = [df.result() for df in df_futures]
+        columns = input.columns
+        matrix = input.T.values.tolist()
 
-        # dfs: List[pl.DataFrame] = await asyncio.gather(*get_df_futs)
-        dfs = [x.select(sorted(x.columns)) for x in dfs]
-        df = pl.concat(dfs)
+        columns_fqn = [str(c) for c in (*columns, CHALK_TS_FEATURE)]
 
-        ts_fqn = next((f for f in output if Feature.from_root_fqn(f).is_feature_time), None)
-        decoded_col_names = self._decode_column_names(
-            df.columns,
-            ts_fqn=None,
-        )
-        df = df.select(list(decoded_col_names.keys()))
-        df = df.rename(dict(decoded_col_names))
-        # The parquet file is all encoded. We gotta decode
-        decoded_stmts = []
-        for col in df.columns:
-            if col == ts_fqn or col == CHALK_TS_FEATURE.root_fqn:
-                decoded_stmts.append(pl.col(col).dt.with_time_zone("UTC"))
-            else:
+        matrix.append([a for a in input_times])
 
-                def _bound_decode(val: Any, col: str = col):
-                    if val is None:
-                        return None
-                    return json.loads(val)
-
-                decoded_stmts.append(
-                    pl.col(col).apply(_bound_decode, return_dtype=FEATURE_CODEC.get_polars_dtype(fqn=col))
+        for col_index, column in enumerate(matrix):
+            for row_index, value in enumerate(column):
+                matrix[col_index][row_index] = FEATURE_CODEC.encode_fqn(
+                    fqn=columns_fqn[col_index],
+                    value=value,
                 )
 
-        df = df.select(decoded_stmts)
-        return df.select(sorted(df.columns))
+        query_input = _OfflineQueryInput(
+            columns=columns_fqn,
+            values=matrix,
+        )
+        request = _ComputeResolverOutputRequest(input=query_input, resolver_fqn=resolver)
+        response = self._request(
+            method="POST", uri="/v1/compute_resolver_output", json=request, response=_ComputeResolverOutputResponse
+        )
+
+        if response.errors is not None and len(response.errors) > 0:
+            raise ChalkOfflineQueryException(message="Failed to execute offline query", errors=response.errors)
+
+        while True:
+            status = self._get_compute_job_status(job_id=response.job_id)
+            if status.is_finished:
+                break
+            time.sleep(0.5)
+
+        return load_dataset(uris=status.urls, version=DatasetVersion.COMPUTE_RESOLVER_OUTPUT_V1, executor=None)
+
+    def _get_compute_job_status(self, job_id: str) -> _GetOfflineQueryJobResponse:
+        return self._request(
+            method="GET", uri=f"/v1/compute_resolver_output/{job_id}", response=_GetOfflineQueryJobResponse
+        )
 
     def _create_offline_query_job(self, request: _CreateOfflineQueryJobRequest):
         response = self._request(
