@@ -2,13 +2,15 @@ from __future__ import annotations
 
 from datetime import datetime
 from enum import Enum
-from typing import Any, List, Mapping, Optional, Protocol, Union, overload
+from typing import TYPE_CHECKING, Any, List, Mapping, Optional, Protocol, Sequence, Union
 
 import pandas as pd
-import polars as pl
 from pydantic import BaseModel
 
-from chalk.features import DataFrame, Feature
+from chalk.features import DataFrame, Feature, Features, FeatureWrapper
+
+if TYPE_CHECKING:
+    import polars as pl
 
 
 class OnlineQueryContext(BaseModel):
@@ -220,9 +222,9 @@ class ChalkAPIClientProtocol(Protocol):
 
     def query(
         self,
-        input: Mapping[Union[str, Feature, Any], Any],
-        output: List[Union[str, Feature, Any]],
-        staleness: Optional[Mapping[Union[str, Feature, Any], str]] = None,
+        input: Mapping[Union[str, Feature, FeatureWrapper, Any], Any],
+        output: Sequence[Union[str, Feature, FeatureWrapper, Any]],
+        staleness: Optional[Mapping[Union[str, Feature, FeatureWrapper, Any], str]] = None,
         context: Optional[OnlineQueryContext] = None,
         preview_deployment_id: Optional[str] = None,
         correlation_id: Optional[str] = None,
@@ -254,7 +256,7 @@ class ChalkAPIClientProtocol(Protocol):
 
     def upload_features(
         self,
-        input: Mapping[Union[str, Feature, Any], Any],
+        input: Mapping[Union[str, Feature, FeatureWrapper, Any], Any],
         context: Optional[OnlineQueryContext] = None,
         preview_deployment_id: Optional[str] = None,
         correlation_id: Optional[str] = None,
@@ -280,90 +282,79 @@ class ChalkAPIClientProtocol(Protocol):
         """
         ...
 
-    @overload
     def get_training_dataframe(
         self,
-        input: Mapping[Union[str, Feature], List[Any]],
-        input_times: List[datetime],
-        output: List[Union[str, Feature]],
-        context: Optional[OfflineQueryContext] = None,
-        dataset: Optional[str] = None,
-        branch: Optional[str] = None,
-    ) -> pd.DataFrame:
-        """
-        Compute feature values from the offline store.
-        See https://docs.chalk.ai/docs/training-client for more information.
-
-        :param input: The features for which there are known values, mapped to those values.
-                      Each element in the list of values represents an observation in line
-                      with the timestamp in `input_times`.
-        :param input_times: A list of the times of the observations from `input`.
-        :param output: The features that you'd like to compute from the inputs.
-        :param context: The environment under which you'd like to query your data.
-        :param dataset: A unique name that if provided will be used to generate and save a dataset
-                        constructed from the list of features computed from the inputs
-        :return: A pandas dataframe with columns equal to the names of the features in output,
-                 and values representing the value of the observation for each input time.
-                 The output maintains the ordering from `input`
-        """
-        ...
-
-    @overload
-    def get_training_dataframe(
-        self,
-        input: Union[pd.DataFrame, pl.DataFrame, DataFrame],
-        input_times: List[datetime],
-        output: List[Union[str, Feature]],
+        input: Union[
+            Mapping[Union[str, Feature], Sequence[Any]],
+            pd.DataFrame,
+            pl.DataFrame,
+            DataFrame,
+        ],
+        input_times: Union[Sequence[datetime], datetime, None] = None,
+        output: Sequence[Union[str, Feature, FeatureWrapper, Any]] = (),
+        required_output: Sequence[Union[str, Feature, FeatureWrapper, Any]] = (),
+        output_ts: bool = True,
         context: Optional[OfflineQueryContext] = None,
         dataset: Optional[str] = None,
         branch: Optional[str] = None,
+        max_samples: Optional[int] = None,
     ) -> pd.DataFrame:
         """
         Compute feature values from the offline store.
         See https://docs.chalk.ai/docs/training-client for more information.
 
         :param input: The features for which there are known values.
-                      Each element in the list of values represents an observation in line
+                      It can be a mapping of features to a list of values for each feature,
+                      or an existing dataframe.
+                      Each element in the dataframe or list of values represents an observation in line
                       with the timestamp in `input_times`.
         :param input_times: A list of the times of the observations from `input`.
-        :param output: The features that you'd like to compute from the inputs.
+        :param output: The features that you'd like to sample, if they exist.
+                       If an output feature was never computed for a sample (row) in the resulting DataFrame,
+                       its value will be ``null``.
+        :param required_output: The features that you'd like to sample and must exist in each resulting row.
+                                Rows where a ``required_output`` was never stored in the offline store will be
+                                skipped. This differs from specifying the feature in ``output``, where instead
+                                the row would be included, but the feature value would be ``null``.
+        :param output_ts: Whether to return the timestamp feature in a column named ``__chalk__.CHALK_TS`` in the
+                          resulting DataFrame.
         :param context: The environment under which you'd like to query your data.
         :param dataset: A unique name that if provided will be used to generate and save a dataset
                         constructed from the list of features computed from the inputs
+        :param max_samples: The maximum number of samples to include in the dataframe. If not specified, then
+                            all samples will be returned.
         :return: A dataframe with columns equal to the names of the features in output,
                  and values representing the value of the observation for each input time.
                  The output maintains the ordering from `input`
         """
         ...
 
-    def get_training_dataframe(
-        self,
-        input: Union[Mapping[Union[str, Feature], Any], pl.DataFrame, pd.DataFrame, DataFrame],
-        input_times: List[datetime],
-        output: List[Union[str, Feature, Any]],
-        context: Optional[OfflineQueryContext] = None,
-        dataset: Optional[str] = None,
-        branch: Optional[str] = None,
-    ) -> pd.DataFrame:
-        """
-        Unified type signature for get_training_dataframe.
-        Refer to the above two functions for information on the types.
-
-        https://docs.chalk.ai/docs/training-client
-        """
-        ...
-
     def sample(
         self,
-        output: List[Union[str, Feature, Any]],
+        output: Sequence[Union[str, Feature, FeatureWrapper, Any]] = (),
+        required_output: Sequence[Union[str, Feature, FeatureWrapper, Any]] = (),
+        output_id: bool = False,
+        output_ts: bool = False,
         max_samples: Optional[int] = None,
+        dataset: Optional[str] = None,
+        branch: Optional[str] = None,
         context: Optional[OfflineQueryContext] = None,
     ) -> pd.DataFrame:
         """
         Get the most recent feature values from the offline store.
         See https://docs.chalk.ai/docs/training-client for more information.
 
-        :param output: The features that you'd like to sample.
+        :param output: The features that you'd like to sample, if they exist.
+                       If an output feature was never computed for a sample (row) in the resulting DataFrame,
+                       its value will be ``null``.
+        :param required_output: The features that you'd like to sample and must exist in each resulting row.
+                                Rows where a ``required_output`` was never stored in the offline store will be
+                                skipped. This differs from specifying the feature in ``output``, where instead
+                                the row would be included, but the feature value would be ``null``.
+        :param output_id: Whether to return the primary key feature in a column named ``__chalk__.__id__`` in the
+                          resulting DataFrame.
+        :param output_ts: Whether to return the timestamp feature in a column named ``__chalk__.CHALK_TS`` in the
+                          resulting DataFrame.
         :param max_samples: The maximum number of rows to return.
         :param context: The environment under which you'd like to query your data.
 
