@@ -11,7 +11,7 @@ from chalk.features.feature_wrapper import FeatureWrapper, unwrap_feature
 from chalk.features.filter import Filter, TimeDelta, after, before
 from chalk.features.hooks import after_all, before_all
 from chalk.features.primary import Primary
-from chalk.features.resolver import Cron, ScheduleOptions, offline, online, sink
+from chalk.features.resolver import Cron, Resolver, ScheduleOptions, offline, online, sink
 from chalk.features.tag import Environments, Tags
 from chalk.utils import MachineType
 
@@ -84,6 +84,9 @@ def owner(f: Any) -> Optional[str]:
     if isinstance(f, (Feature, FeatureWrapper)):
         feature = unwrap_feature(f)
         return feature.owner
+    if isinstance(f, Resolver):
+        return f.owner
+
     raise TypeError(f"Could not determine the owner of {f} as it is neither a Feature or Feature Set")
 
 
