@@ -1,12 +1,16 @@
 from __future__ import annotations
 
+import base64
 import collections.abc
 import datetime
 import enum
 import functools
 import inspect
+import json
 import operator
+import os
 import pathlib
+import typing
 import warnings
 from typing import (
     TYPE_CHECKING,
@@ -27,9 +31,11 @@ from typing import (
     overload,
 )
 
-import polars as pl
+import pydantic
 from pydantic import BaseModel
 
+from chalk.utils.missing_dependency import missing_dependency_exception
+
 try:
     import zoneinfo
 except ImportError:
@@ -46,9 +52,11 @@ from chalk.utils.collections import ensure_tuple, get_unique_item
 
 if TYPE_CHECKING:
     import pandas as pd
+    import polars as pl
     import pyarrow
 
     from chalk.features.feature_set import Features
+    from chalk.sql import ChalkQueryProtocol, FinalizedChalkQuery, StringChalkQueryProtocol
 
 else:
     try:
@@ -74,7 +82,12 @@ def _iso_parse(x: str, expected_tz: datetime.tzinfo | None) -> datetime.datetime
 
 
 class DataFrameMeta(type):
-    def __getitem__(cls, item: Any) -> Type[DataFrame]:
+    # Annotating __getitem__ return a union of DataFrame, query, or string to support resolvers that return sql queries
+    def __getitem__(
+        cls, item: Any
+    ) -> Union[
+        Type[DataFrame], Type[ChalkQueryProtocol], Type[StringChalkQueryProtocol], Type[FinalizedChalkQuery], Type[str]
+    ]:
         from chalk.features.feature_set import Features
 
         # leaving untyped as we type the individual features as their object type
@@ -201,7 +214,7 @@ class DataFrameMeta(type):
         return get_unique_item(namespaces, f"dataframe {cls.__name__} column namespaces")
 
 
-class DataFrame(metaclass=DataFrameMeta):
+class DataFrameImpl(metaclass=DataFrameMeta):
     filters: ClassVar[Tuple[Filter, ...]] = ()
     columns: Tuple[Feature, ...]  # set via a @property on the metaclass
     __columns__: ClassVar[Tuple[Feature, ...]] = ()
@@ -233,6 +246,11 @@ class DataFrame(metaclass=DataFrameMeta):
         """
         from chalk.features.feature_set import Features, FeatureSetBase
 
+        try:
+            import polars as pl
+        except ImportError:
+            raise missing_dependency_exception("chalkpy[polars]")
+
         self._pydantic_model = pydantic_model
 
         # Typing the keys of ``data`` as Any, as {FeatureCls.item: x} would be typed as the underlying annotation of the features cls
@@ -292,34 +310,41 @@ class DataFrame(metaclass=DataFrameMeta):
                 # str(Feature) and str(FeatureWrapper) return the root fqns
                 new_data_dict: Dict[str, Sequence[Any] | pl.Series] = {}
                 for (k, v) in data.items():
+                    k_str = str(k)
                     v = [_preprocess_element(x) for x in v]
 
                     if convert_dtypes:
-                        polars_dtype = FEATURE_CODEC.get_polars_dtype(str(k))
-                        elements_as_series = _generate_empty_series_for_dtype(str(k), polars_dtype, 0)
-                        null_element = _generate_empty_series_for_dtype(str(k), polars_dtype, 1)
+                        polars_dtype = FEATURE_CODEC.get_polars_dtype(k_str)
+                        elements_as_series = _generate_empty_series_for_dtype(k_str, polars_dtype, 0)
+                        null_element = _generate_empty_series_for_dtype(k_str, polars_dtype, 1)
                         for i, element in enumerate(v):
-                            if element is None or isinstance(element, (list, tuple)) and len(element) == 0:
+                            if element is None:
                                 elements_as_series.append(null_element)
+                            elif isinstance(element, (list, tuple)) and len(element) == 0:
+                                assert isinstance(polars_dtype, pl.List)
+                                assert polars_dtype.inner is not None
+                                elements_as_series.append(
+                                    pl.Series(values=[_generate_empty_series_for_dtype(k_str, polars_dtype.inner, 0)])
+                                )
+                            elif isinstance(element, pydantic.BaseModel):
+                                n = FEATURE_CODEC._to_polars_struct(fqn=k_str, value=element)
+                                elements_as_series.append(n)
                             else:
-                                element_as_series = pl.Series(name=str(k), values=[element])
+                                element_as_series = pl.Series(name=k_str, values=[element])
                                 if element_as_series.dtype != polars_dtype:
-
                                     if _polars_dtype_contains_struct(polars_dtype):
                                         # Cannot cast to a struct type. Instead, will error, so the user can ensure the underlying
                                         # dictionaries / dataclasses are of the correct type
                                         raise TypeError(
                                             (
-                                                f"Expected field `{str(k)}` to have dtype `{str(polars_dtype)}`; got dtype `{str(element_as_series.dtype)}`"
+                                                f"Expected field `{k_str}` to have dtype `{str(polars_dtype)}`; got dtype `{str(element_as_series.dtype)}`"
                                                 f" for element {element} at index {i}"
                                             )
                                         )
                                     else:
                                         try:
-                                            if (
-                                                isinstance(element_as_series.dtype, type)
-                                                and issubclass(element_as_series.dtype, pl.Utf8)
-                                                and isinstance(polars_dtype, pl.Datetime)
+                                            if element_as_series.dtype == pl.Utf8 and isinstance(
+                                                polars_dtype, pl.Datetime
                                             ):
                                                 if polars_dtype.tz is None:
                                                     element_as_series = element_as_series.apply(
@@ -330,12 +355,16 @@ class DataFrame(metaclass=DataFrameMeta):
                                                     element_as_series = element_as_series.apply(
                                                         lambda x: None if x is None else _iso_parse(x, tzinfo),
                                                     ).dt.with_time_zone(polars_dtype.tz)
+                                            elif element_as_series.dtype == pl.Utf8 and polars_dtype == pl.Date:
+                                                element_as_series = element_as_series.apply(
+                                                    lambda x: None if x is None else _iso_parse(x, None),
+                                                ).cast(pl.Date)
                                             else:
                                                 element_as_series = element_as_series.cast(polars_dtype)
                                         except (pl.ComputeError, pl.InvalidOperationError) as e:
                                             # If operating on a lazy frame, we won't get the exception until we call .compute()
                                             raise TypeError(
-                                                f"Feature `{str(k)}` at index {i} with value '{element}' could not be converted to dtype `{polars_dtype.string_repr()}`"
+                                                f"Feature `{k_str}` at index {i} with value '{element}' could not be converted to dtype `{polars_dtype.string_repr()}`"
                                             ) from e
                                 elements_as_series.append(element_as_series)
                         elements_as_series = elements_as_series.rechunk(True)
@@ -366,7 +395,16 @@ class DataFrame(metaclass=DataFrameMeta):
                 expected_dtype = FEATURE_CODEC.get_polars_dtype(root_fqn)
                 actual_dtype = underlying.schema[root_fqn]
                 if actual_dtype != expected_dtype:
-                    if _polars_dtype_contains_struct(expected_dtype):
+                    if isinstance(underlying, pl.LazyFrame):
+                        underlying = underlying.collect()
+                    if len(underlying) == underlying.get_column(root_fqn).null_count():
+                        # If all values are null, then replace with a null series of the correct datatype
+                        # It's quite possible that the original column will have an incorrect dtype if all values are null,
+                        # since there was no data to infer the correct dtype from
+                        underlying = underlying.with_column(
+                            _generate_empty_series_for_dtype(root_fqn, expected_dtype, len(underlying))
+                        )
+                    elif _polars_dtype_contains_struct(expected_dtype):
                         # Cannot cast to a struct type. Instead, will error, so the user can ensure the underlying
                         # dictionaries / dataclasses are of the correct type
                         raise TypeError(
@@ -394,6 +432,14 @@ class DataFrame(metaclass=DataFrameMeta):
                                         .dt.with_time_zone(expected_dtype.tz)
                                     )
                                 underlying = underlying.with_columns([col])
+                            elif actual_dtype == pl.Utf8 and expected_dtype == pl.Date:
+                                col = (
+                                    pl.col(root_fqn)
+                                    .apply(
+                                        lambda x: None if x is None else _iso_parse(cast(str, x), None),
+                                    )
+                                    .cast(pl.Date)
+                                )
                             else:
                                 underlying = underlying.with_columns([pl.col(root_fqn).cast(expected_dtype)])
                         except pl.ComputeError as e:
@@ -456,21 +502,50 @@ class DataFrame(metaclass=DataFrameMeta):
         df = cls(features_seq)
         return df
 
+    @classmethod
+    def _get_storage_options(cls) -> typing.Optional[Dict[str, Any]]:
+        if os.getenv("GCP_INTEGRATION_CREDENTIALS_B64") is not None:
+            try:
+                from google.oauth2 import service_account
+            except ImportError:
+                raise missing_dependency_exception("google-auth")
+            token = service_account.Credentials.from_service_account_info(
+                json.loads(base64.b64decode(os.getenv("GCP_INTEGRATION_CREDENTIALS_B64"))),
+                scopes=[
+                    "https://www.googleapis.com/auth/cloud-platform",
+                    "https://www.googleapis.com/auth/userinfo.email",
+                ],
+            )
+            return {"token": token}
+        else:
+            return None
+
     @classmethod
     def read_parquet(
         cls: Type[TDataFrame],
         path: Union[str, pathlib.Path],
-        columns: Optional[Union[List[int], List[str], Dict[str, Union[str, Feature, FeatureWrapper, Any]]]] = None,
+        columns: Optional[
+            Union[
+                Dict[str, Union[str, Feature, Any]],
+                Dict[int, Union[str, Feature, Any]],
+            ]
+        ] = None,
     ) -> TDataFrame:
-        if isinstance(columns, dict):
-            columns_to_read = list(columns.keys())
-            column_map = {k: str(v) for (k, v) in columns.items()}
-        else:
-            columns_to_read = columns
-            column_map = None
-        data = pl.read_parquet(path, columns_to_read)
-        if column_map is not None:
-            data = data.rename(column_map)
+        try:
+            import polars as pl
+        except ImportError:
+            raise missing_dependency_exception("chalkpy[polars]")
+
+        if not isinstance(columns, dict):
+            raise ValueError(f"Invalid column mapping. Received '{type(columns).__name__}'; expected dict")
+
+        data = pl.read_parquet(
+            source=path,
+            columns=list(columns.keys()),
+            storage_options=DataFrame._get_storage_options(),
+        )
+        col_values = [str(x) for x in columns.values()]
+        data = data.rename({c: col_values[i] for i, c in enumerate(data.columns)})
         return cls(data)
 
     @classmethod
@@ -478,17 +553,31 @@ class DataFrame(metaclass=DataFrameMeta):
         cls: Type[TDataFrame],
         path: Union[str, pathlib.Path],
         has_header: bool = True,
-        columns: Optional[Union[List[int], List[str], Dict[str, Union[str, Feature, FeatureWrapper, Any]]]] = None,
+        columns: Optional[
+            Union[
+                Dict[str, Union[str, Feature, Any]],
+                Dict[int, Union[str, Feature, Any]],
+            ]
+        ] = None,
     ) -> TDataFrame:
-        if isinstance(columns, dict):
-            columns_to_read = list(columns.keys())
-            column_map = {k: str(v) for (k, v) in columns.items()}
-        else:
-            columns_to_read = columns
-            column_map = None
-        data = pl.read_csv(path, has_header, columns_to_read)
-        if column_map is not None:
-            data = data.rename(column_map)
+        try:
+            import polars as pl
+        except ImportError:
+            raise missing_dependency_exception("chalkpy[polars]")
+
+        if not isinstance(columns, dict):
+            raise ValueError(f"Invalid column mapping. Received '{type(columns).__name__}'; expected dict")
+
+        col_values = [str(x) for x in columns.values()]
+        dtypes = [FEATURE_CODEC.get_polars_dtype(fqn) for fqn in col_values]
+        data = pl.read_csv(
+            file=path,
+            has_header=has_header,
+            columns=list(columns.keys()),
+            dtypes=dtypes,
+            storage_options=DataFrame._get_storage_options(),
+        )
+        data = data.rename({c: col_values[i] for i, c in enumerate(data.columns)})
         return cls(data)
 
     #############
@@ -530,6 +619,8 @@ class DataFrame(metaclass=DataFrameMeta):
 
     def any(self):
         """Returns whether any of the values in the dataframe are truthy. Requires the dataframe to only contain boolean values."""
+        import polars as pl
+
         if len(self) == 0:
             return False
         if not all(isinstance(x, type) and issubclass(x, pl.Boolean) for x in self._underlying.dtypes):
@@ -539,6 +630,8 @@ class DataFrame(metaclass=DataFrameMeta):
 
     def all(self):
         """Returns whether all of the values in the dataframe are truthy. Requires the dataframe to only contain boolean values."""
+        import polars as pl
+
         if len(self) == 0:
             return True
         if not all(isinstance(x, type) and issubclass(x, pl.Boolean) for x in self._underlying.dtypes):
@@ -596,6 +689,8 @@ class DataFrame(metaclass=DataFrameMeta):
         other: Union[DataFrame, pl.DataFrame, pd.DataFrame, Any],
         convert_dtypes: bool,
     ):
+        import polars as pl
+
         materialized = self._materialize()
         if isinstance(other, DataFrame):
             other = other.to_polars()
@@ -627,9 +722,13 @@ class DataFrame(metaclass=DataFrameMeta):
     ):
         # There's a bug in the default polars implementation for comparisons -- see
         # https://github.com/pola-rs/polars/issues/5870
+        import polars as pl
+
         materialized = self._materialize()
-        if materialized.columns != other.columns:
-            raise ValueError("DataFrame columns do not match")
+        if set(materialized.columns) != set(other.columns):
+            raise ValueError(f"DataFrame columns do not match. {materialized.columns} != {other.columns}")
+        # Put the columns in the same order
+        other = other.select(materialized.columns)
         if materialized.shape != other.shape:
             raise ValueError("DataFrame dimensions do not match")
 
@@ -833,7 +932,7 @@ class DataFrame(metaclass=DataFrameMeta):
                     raise KeyError(f"Attempt to select missing columns {missing_cols}")
 
                 df = df.select(projections)
-            return DataFrame(df, pydantic_model=self)
+            return DataFrame(df, pydantic_model=self._pydantic_model)
         else:
             # Otherwise, use the standard polars selection format
             # Must materialize the dataframe to use __getitem__
@@ -851,6 +950,14 @@ class DataFrame(metaclass=DataFrameMeta):
         return DataFrame(self._underlying.slice(offset, length))
 
 
+# Hack to get VSCode/Pylance/Pyright to type DataFrame as Type[DataFrameImpl]
+# but IntelliJ to type it as Type[Any]
+# Vscode can parse through literal dicts; IntelliJ can't
+_dummy_dict = {"0": DataFrameImpl}
+
+DataFrame = _dummy_dict["0"]
+
+
 def _feature_type_or_value(e: Union[Feature, FeatureWrapper]):
     if isinstance(e, FeatureWrapper):
         e = unwrap_feature(e)
@@ -859,6 +966,8 @@ def _feature_type_or_value(e: Union[Feature, FeatureWrapper]):
 
 def _polars_dtype_contains_struct(dtype: pl.DataType | Type[pl.DataType]):
     """Returns whether the dtype contains a (potentially nested) struct"""
+    import polars as pl
+
     if isinstance(dtype, pl.Struct) or (isinstance(dtype, type) and issubclass(dtype, pl.Struct)):
         return True
     if isinstance(dtype, pl.List):
@@ -912,6 +1021,8 @@ def _polars_is_eq(
     """Compare a column with another column or literal value, including possibly a struct.
     Polars does not permit equality comparisons on structs directly. Instead, must compare field by field, potentially recursively.
     """
+    import polars as pl
+
     if rhs_dtype is not None:
         if lhs_dtype is None:
             # Swap the columns
@@ -961,6 +1072,8 @@ def _polars_is_eq(
 
 def _polars_is_in(lhs: pl.Expr, rhs: Iterable, lhs_dtype: Union[Type[pl.DataType], pl.DataType]):
     """Filter for where the lhs is in the RHS. The RHS must be a literal collection."""
+    import polars as pl
+
     if isinstance(lhs_dtype, pl.Struct):
         # Assert equality field by field
         filters = []
@@ -989,6 +1102,8 @@ def _polars_is_in(lhs: pl.Expr, rhs: Iterable, lhs_dtype: Union[Type[pl.DataType
 
 
 def _coerce_value_to_dtype(val: Any, dtype: Union[pl.DataType, Type[pl.DataType]]):
+    import polars as pl
+
     # For the most part, we don't need to do any manual coercion -- polars will handle that for us.
     # Only need to pay attention to enums, which could be stored as the underlying type, or as an object
     if isinstance(val, collections.abc.Iterable) and not isinstance(val, str):
@@ -1016,6 +1131,8 @@ def _convert_filter_to_pl_expr(
     Returns:
         A series of boolean values that can be used to select the rows where the filter is truthy
     """
+    import polars as pl
+
     # Passing `now` in explicitly instead of using datetime.datetime.now() so that multiple filters
     # relying on relative timestamps (e.g. before, after) will have the same "now" time.
     if f.operation == "not":
@@ -1109,6 +1226,8 @@ def _generate_empty_series_for_dtype(name: str, dtype: Union[Type[pl.DataType],
 
     Unlike the ``pl.Series`` constructor, this function can handle struct dtypes.
     """
+    import polars as pl
+
     if isinstance(dtype, pl.Struct):
         # Struct dtypes cannot be specified in the pl.Series constructor.
         # Instead, create a dataframe, then call .to_struct() on it
@@ -1116,18 +1235,20 @@ def _generate_empty_series_for_dtype(name: str, dtype: Union[Type[pl.DataType],
         data = {f.name: _generate_empty_series_for_dtype(f.name, f.dtype, length) for f in dtype.fields}
         temp_df = pl.DataFrame(data)
         return temp_df.to_struct(name)
-    if isinstance(dtype, pl.List):
+    if isinstance(dtype, pl.List) and isinstance(dtype.inner, pl.Struct):
+        # The pl.Series constructor does not respect nested data types
+        # So, we'll manually build a dataframe with the list set to the correct type,
+        # set it to None, and then select just the list
         assert dtype.inner is not None
         data = {name: _generate_empty_series_for_dtype(name, dtype.inner, 0)}
         temp_df = pl.DataFrame(data)
-        list_of_struct_series = temp_df.select(pl.col(name).reshape((length, -1))).get_column(name)
+        df_with_list = temp_df.select(pl.col(name).reshape((length, -1)))
+        df_with_list = df_with_list.select(pl.when(True).then(None).otherwise(pl.col(name)).alias(name))
+        list_of_struct_series = df_with_list.select(pl.col(name).reshape((length, -1))).get_column(name)
         return list_of_struct_series
     return pl.Series(name, dtype=dtype, values=([None] * length))
 
 
-# list [[], [None, None]]
-
-
 def _preprocess_element(element: Any):
     if isinstance(element, enum.Enum):
         element = element.value
