@@ -2,11 +2,11 @@ from __future__ import annotations
 
 import copy
 from collections.abc import Iterable
-from typing import TYPE_CHECKING, Any, Optional, TypeVar, Union, cast
+from typing import TYPE_CHECKING, Any, Optional, TypeVar, Union
 
 from chalk.features.filter import Filter
 from chalk.serialization.parsed_annotation import ParsedAnnotation
-from chalk.streams import Windowed, get_name_with_duration
+from chalk.streams import get_name_with_duration
 from chalk.utils.collections import ensure_tuple
 
 if TYPE_CHECKING:
@@ -103,8 +103,10 @@ class FeatureWrapper:
         if not hasattr(parent, desired_attribute_name):
             formatted_window_durations = [f"'{x}s'" for x in self._chalk_feature.window_durations]
             raise TypeError(
-                f"Unsupported window duration '{window}' for '{self._chalk_feature.root_fqn}'. "
-                f"Durations {', '.join(formatted_window_durations)} are supported."
+                (
+                    f"Unsupported window duration '{window}' for '{self._chalk_feature.root_fqn}'. "
+                    f"Durations {', '.join(formatted_window_durations)} are supported."
+                )
             )
         return getattr(parent, desired_attribute_name)
 
