@@ -23,7 +23,6 @@ from typing import (
     overload,
 )
 
-import polars
 import pyarrow
 from pydantic import BaseModel
 from typing_extensions import ParamSpec, get_args, get_origin
@@ -52,7 +51,7 @@ from chalk.utils.duration import Duration, ScheduleOptions
 from chalk.utils.environment_parsing import env_var_bool
 
 if TYPE_CHECKING:
-    from chalk.sql.base.protocols import BaseSQLSourceProtocol
+    from chalk.sql import BaseSQLSourceProtocol
 
 
 T = TypeVar("T")
@@ -100,7 +99,7 @@ def _flatten_features(output: Optional[Type[Features]]) -> Sequence[Feature]:
         return features
 
 
-class Resolver(Protocol[P, T]):
+class Resolver(Generic[P, T]):
     function_definition: str
     fqn: str
     filename: str
@@ -111,6 +110,7 @@ class Resolver(Protocol[P, T]):
     tags: Optional[List[str]]
     max_staleness: Optional[Duration]
     machine_type: Optional[MachineType]
+    owner: Optional[str]
     state: Optional[StateDescriptor]
     default_args: List[Optional[ResolverArgErrorHandler]]
 
@@ -121,7 +121,7 @@ class Resolver(Protocol[P, T]):
         return hash(self.fqn)
 
     def __call__(self, *args: P.args, **kwds: P.kwargs) -> T:
-        return self.fn(*args, **kwds)
+        ...
 
     @property
     def __name__(self):
@@ -129,12 +129,17 @@ class Resolver(Protocol[P, T]):
 
 
 def _process_call(result: T, *, declared_output: Optional[Type[Features]]) -> T:
-    from chalk.sql import StringChalkQuery
-    from chalk.sql.integrations.chalk_query import ChalkQuery
+    # __call__ is defined to support userland code that invokes a resolver
+    # as if it is a normal python function
+    # If the user returns a ChalkQuery, then we'll want to automatically execute it
+    from chalk.sql import FinalizedChalkQuery
+    from chalk.sql._internal.chalk_query import ChalkQuery
+    from chalk.sql._internal.string_chalk_query import StringChalkQuery
 
     if isinstance(result, (ChalkQuery, StringChalkQuery)):
-        return result.execute_internal(expected_features=_flatten_features(declared_output))
-
+        result = result.all()
+    if isinstance(result, FinalizedChalkQuery):
+        result = result.execute(_flatten_features(declared_output))
     return result
 
 
@@ -163,8 +168,10 @@ class SinkResolver(Resolver[P, T]):
         debounce: Optional[Duration],
         max_delay: Optional[Duration],
         upsert: Optional[bool],
+        owner: Optional[str],
         integration: Optional[BaseSQLSourceProtocol] = None,
     ):
+        self.owner = owner
         self.function_definition = function_definition
         self.fqn = fqn
         self.filename = filename
@@ -217,6 +224,7 @@ class OnlineResolver(Resolver[P, T]):
         when: Optional[Filter],
         state: Optional[StateDescriptor],
         default_args: List[Optional[ResolverArgErrorHandler]],
+        owner: Optional[str],
     ):
         self.function_definition = function_definition
         self.fqn = fqn
@@ -236,6 +244,7 @@ class OnlineResolver(Resolver[P, T]):
         self.when = when
         self.state = state
         self.default_args = default_args
+        self.owner = owner
 
     def __repr__(self):
         return f"OnlineResolver(name={self.fqn})"
@@ -265,6 +274,7 @@ class OfflineResolver(Resolver[P, T]):
         state: Optional[StateDescriptor],
         when: Optional[Filter],
         default_args: List[Optional[ResolverArgErrorHandler]],
+        owner: Optional[str],
     ):
         self.when = when
         self.function_definition = function_definition
@@ -284,9 +294,11 @@ class OfflineResolver(Resolver[P, T]):
         self.machine_type = machine_type
         self.state = state
         self.default_args = default_args
+        self.owner = owner
 
-    def __repr__(self):
-        return f"OfflineResolver(name={self.fqn})"
+
+def __repr__(self):
+    return f"OfflineResolver(name={self.fqn})"
 
 
 @dataclasses.dataclass(frozen=True)
@@ -404,7 +416,7 @@ def parse_function(
 
     for i, (arg_name, parameter) in enumerate(sig.parameters.items()):
         bad_input = lambda: ValueError(
-            f"Resolver inputs must be Features or State. Received {str(inputs[i])} for argument '{arg_name}'."
+            f"Resolver inputs must be Features or State. Received {str(inputs[i])} for argument '{arg_name}' for '{fqn}'.\n"
         )
         arg = inputs[i]
 
@@ -502,6 +514,7 @@ def online(
     cron: Optional[Union[ScheduleOptions, Cron]] = None,
     machine_type: Optional[MachineType] = None,
     when: Optional[Any] = None,
+    owner: Optional[str] = None,
 ) -> Callable[[Callable[P, T]], Callable[P, T]]:
     ...
 
@@ -523,6 +536,7 @@ def online(
     cron: Optional[Union[ScheduleOptions, Cron]] = None,
     machine_type: Optional[MachineType] = None,
     when: Optional[Any] = None,
+    owner: Optional[str] = None,
 ) -> Union[Callable[[Callable[P, T]], Callable[P, T]], Callable[P, T]]:
     """
     :param fn: The function that you're decorating as a resolver.
@@ -541,6 +555,10 @@ def online(
 
         Read more at https://docs.chalk.ai/docs/resolver-environments
 
+    :param owner: Owner allows you to specify an individual or team
+        who is responsible for this resolver. The Chalk Dashboard
+        will display this field, and alerts can be routed to owners.
+
     :param tags: Tags allow you to scope requests within an
         environment. Both tags and environment need to match for a
         resolver to be a candidate to execute.
@@ -604,6 +622,7 @@ def online(
             cron=cron,
             machine_type=machine_type,
             when=when,
+            owner=owner,
             state=parsed.state,
             default_args=parsed.default_args,
         )
@@ -627,6 +646,7 @@ def offline(
     cron: Optional[Union[ScheduleOptions, Cron]] = None,
     machine_type: Optional[MachineType] = None,
     when: Optional[Any] = None,
+    owner: Optional[str] = None,
 ) -> Callable[[Callable[P, T]], Callable[P, T]]:
     ...
 
@@ -648,6 +668,7 @@ def offline(
     cron: Union[ScheduleOptions, Cron] = None,
     machine_type: Optional[MachineType] = None,
     when: Optional[Any] = None,
+    owner: Optional[str] = None,
 ) -> Union[Callable[[Callable[P, T]], Callable[P, T]], Callable[P, T]]:
     """
     :param fn: The function that you're decorating as a resolver.
@@ -666,6 +687,10 @@ def offline(
 
         Read more at https://docs.chalk.ai/docs/resolver-environments
 
+    :param owner: Owner allows you to specify an individual or team
+        who is responsible for this resolver. The Chalk Dashboard
+        will display this field, and alerts can be routed to owners.
+
     :param tags: Tags allow you to scope requests within an
         environment. Both tags and environment need to match for a
         resolver to be a candidate to execute.
@@ -729,12 +754,13 @@ def offline(
             machine_type=machine_type,
             state=parsed.state,
             when=when,
+            owner=owner,
             default_args=parsed.default_args,
         )
         Resolver.registry.append(resolver)
         if Resolver.hook:
             Resolver.hook(resolver)
-        return fn
+        return resolver
 
     return decorator(fn) if fn else decorator
 
@@ -783,26 +809,25 @@ def sink(
     def decorator(fn: Callable[P, T], cf: str = caller_filename):
         parsed = parse_function(fn, caller_globals, caller_locals, ignore_return=True)
         assert parsed.output is not None
-        SinkResolver.registry.append(
-            SinkResolver(
-                filename=cf,
-                function_definition=parsed.function_definition,
-                fqn=parsed.fqn,
-                doc=parsed.doc,
-                inputs=parsed.inputs,
-                output=parsed.output,
-                fn=fn,
-                environment=None if environment is None else list(ensure_tuple(environment)),
-                tags=None if tags is None else list(ensure_tuple(tags)),
-                machine_type=machine_type,
-                buffer_size=buffer_size,
-                debounce=debounce,
-                max_delay=max_delay,
-                upsert=upsert,
-                integration=integration,
-            )
+        resolver = SinkResolver(
+            filename=cf,
+            function_definition=parsed.function_definition,
+            fqn=parsed.fqn,
+            doc=parsed.doc,
+            inputs=parsed.inputs,
+            output=parsed.output,
+            fn=fn,
+            environment=None if environment is None else list(ensure_tuple(environment)),
+            tags=None if tags is None else list(ensure_tuple(tags)),
+            machine_type=machine_type,
+            buffer_size=buffer_size,
+            debounce=debounce,
+            max_delay=max_delay,
+            upsert=upsert,
+            integration=integration,
         )
-        return fn
+        SinkResolver.registry.append(resolver)
+        return resolver
 
     return decorator(fn) if fn else decorator
 
@@ -937,10 +962,15 @@ def _parse_stream_resolver_param(
         if _is_stream_resolver_body_type(item_typ):
             return StreamResolverParamMessageWindow(name=param.name, typ=annotation)
 
-    if is_windowed_resolver and isclass(annotation) and issubclass(annotation, (pyarrow.Table, polars.DataFrame)):
-        return StreamResolverParamMessageWindow(name=param.name, typ=annotation)
-
-    if is_windowed_resolver and isclass(annotation) and issubclass(annotation, DataFrame):
+    if (
+        is_windowed_resolver
+        and isclass(annotation)
+        and (
+            issubclass(annotation, pyarrow.Table)
+            or annotation.__name__ in ("DataFrame", "DataFrameImpl", "SubclassedDataFrame")
+        )
+    ):
+        # Using string comparison as polars may not be installed
         return StreamResolverParamMessageWindow(name=param.name, typ=annotation)
 
     raise ValueError(
