@@ -4,11 +4,10 @@ import os
 import sys
 import traceback
 from pathlib import Path
-from typing import List
-
-from pydantic import BaseModel
+from typing import Callable, List
 
 from chalk.gitignore.gitignore_parser import parse_gitignore
+from chalk.parsed.duplicate_input_gql import FailedImport
 from chalk.utils.log_with_context import get_logger
 from chalk.utils.paths import get_directory_root
 
@@ -23,12 +22,6 @@ def py_path_to_module(path: Path, repo_root: Path) -> str:
     return str(p)[: -len(".py")].replace("./", "").replace("/", ".")
 
 
-class FailedImport(BaseModel):
-    filename: str
-    module: str
-    traceback: str
-
-
 def import_all_python_files() -> List[FailedImport]:
     project_root = get_directory_root()
     if project_root is None:
@@ -55,15 +48,18 @@ def _search_recursively_for_file(base: Path, filename: str) -> List[Path]:
         base = parent
 
 
+def _is_relevant(x: Path, matching_functions: List[Callable[[Path], bool]]):
+    return all((not match(x) for match in matching_functions))
+
+
 def import_all_python_files_from_dir(project_root: Path) -> List[FailedImport]:
     project_root = project_root.absolute()
-    matching_functions = []
+    matching_functions: List[Callable[[Path], bool]] = []
     matching_functions.extend(parse_gitignore(str(x)) for x in _search_recursively_for_file(project_root, ".gitignore"))
     matching_functions.extend(
         parse_gitignore(str(x)) for x in _search_recursively_for_file(project_root, ".chalkignore")
     )
-
-    is_relevant = lambda x: all((not match(x) for match in matching_functions))
+    matching_functions.append(lambda x: x.name == "setup.py")
 
     cwd = os.getcwd()
     os.chdir(project_root)
@@ -87,7 +83,11 @@ def import_all_python_files_from_dir(project_root: Path) -> List[FailedImport]:
 
     errors: List[FailedImport] = []
     for (module_path, filename) in module_paths:
-        if module_path.startswith(".eggs") or module_path.startswith("venv") or not is_relevant(filename):
+        if (
+            module_path.startswith(".eggs")
+            or module_path.startswith("venv")
+            or not _is_relevant(filename, matching_functions)
+        ):
             continue
 
         try:
