@@ -8,20 +8,25 @@ from typing import TYPE_CHECKING, Any, List, Type, Union, cast, get_type_hints
 
 import cattrs
 import pendulum
-import polars as pl
+import pyarrow
+import pyarrow as pa
 import pydantic
 import pytz
 from dateutil import parser
 from pendulum.tz.timezone import Timezone
-from polars.internals.construction import is_namedtuple
+from pydantic import BaseModel
+from pydantic.fields import ModelField
 from typing_extensions import get_args, get_origin
 
 from chalk.serialization.parsed_annotation import ParsedAnnotation
 from chalk.streams import Windowed
 from chalk.utils.enum import get_enum_value_type
 from chalk.utils.log_with_context import get_logger
+from chalk.utils.missing_dependency import missing_dependency_exception
 
 if TYPE_CHECKING:
+    import polars as pl
+
     from chalk.features import Feature
 
 try:
@@ -45,16 +50,19 @@ _log = get_logger(__name__)
 _NoneType = type(None)
 
 
-def _unwrap_optional_if_needed(typ: Type):
+def _unwrap_optional_if_needed(typ: Any):
     if get_origin(typ) == Union and len(get_args(typ)) == 2:
         has_none = any(d == _NoneType for d in get_args(typ))
         if has_none:
-            return next((m for m in get_args(typ) if m != _NoneType), None)
+            return next(m for m in get_args(typ) if m != _NoneType)
 
     return typ
 
 
 def is_numeric(dtype: Union[pl.DataType, Type[pl.DataType]]) -> bool:
+    # This function is used only by the engine
+    import polars as pl
+
     return dtype in (
         pl.Float32,
         pl.Float64,
@@ -69,6 +77,14 @@ def is_numeric(dtype: Union[pl.DataType, Type[pl.DataType]]) -> bool:
     )
 
 
+def _is_namedtuple(value: Any, annotated: bool = False) -> bool:
+    """Infer whether value is a NamedTuple."""
+    # From https://github.com/pola-rs/polars/blob/5f3e332fb2a653064f083b02949c527e0ec0afda/py-polars/polars/internals/construction.py#L78
+    if all(hasattr(value, attr) for attr in ("_fields", "_field_defaults", "_replace")):
+        return len(value.__annotations__) == len(value._fields) if annotated else True
+    return False
+
+
 class FeatureCodec:
     def __init__(self):
         self.converter = cattrs.Converter()
@@ -127,6 +143,24 @@ class FeatureCodec:
 
         return self.encode(Feature.from_root_fqn(fqn), value)
 
+    def _to_polars_struct(self, fqn: str, value: BaseModel):
+        try:
+            import polars as pl
+        except ImportError:
+            raise missing_dependency_exception("chalkpy[polars]")
+
+        return pl.DataFrame(
+            data={
+                field.alias or field.name: [getattr(value, field.alias or field.name)]
+                for field in value.__fields__.values()
+            },
+            columns={
+                field.alias
+                or field.name: self._pa_to_pl(self._get_pa_dtype_for_type(field.type_, field.alias or field.name))
+                for field in value.__fields__.values()
+            },
+        ).to_struct(fqn)
+
     def _default_decode_value(
         self,
         feature: Feature,
@@ -196,7 +230,12 @@ class FeatureCodec:
     ):
         from chalk.features import Feature
 
-        return self.decode(Feature.from_root_fqn(fqn), value)
+        try:
+            f = Feature.from_root_fqn(fqn)
+        except:
+            return value
+
+        return self.decode(f, value)
 
     def get_pandas_dtype(self, fqn: str) -> str:
         from chalk.features import Feature
@@ -226,64 +265,121 @@ class FeatureCodec:
         return "object"
 
     def get_polars_dtype(self, fqn: str) -> Union[pl.DataType, Type[pl.DataType]]:
+        try:
+            import polars as pl
+        except ImportError:
+            raise missing_dependency_exception("chalkpy[polars]")
+
+        # BEGIN This should be replaced with a structification
         from chalk.features import Feature
 
         feature = Feature.from_root_fqn(fqn)
-        if feature.polars_dtype is not None:
-            return feature.polars_dtype
+        # END structification workaround
+
+        if feature.is_has_one or feature.is_has_many:
+            return pl.Object
+        pa_type = self.get_pyarrow_dtype(fqn)
+
+        try:
+            return self._pa_to_pl(pa_type)
+        except Exception as exc:
+            raise ValueError(
+                f"Failed to determined type for feature {fqn}. Please set the pyarrow_dtype attribute directly."
+            ) from exc
+
+    def _pa_to_pl(self, pa_type: pa.DataType) -> Union[pl.DataType, Type[pl.DataType]]:
+        try:
+            import polars as pl
+        except ImportError:
+            raise missing_dependency_exception("chalkpy[polars]")
+
+        """May throw"""
+        # polars coerces pa.Date to pl.DateTime
+        if pa_type == pa.date64():
+            return pl.Date
+
+        # To convert the dtypes, we'll create an empty pyarrow table, and then convert that to polars, and get the dtype
+        pa_table = pa.Table.from_arrays(arrays=[[]], schema=pa.schema(fields={"col": pa_type}))
+
+        df = pl.from_arrow(pa_table)
+        assert isinstance(df, pl.DataFrame)
+        return df.dtypes[0]
+
+    def get_pyarrow_dtype(self, fqn: str) -> pa.DataType:
+        from chalk.features import Feature
+
+        feature = Feature.from_root_fqn(fqn)
+        if feature.pyarrow_dtype is not None:
+            return feature.pyarrow_dtype
         typ = feature.typ
         assert typ is not None, "typ should be specified"
         underlying = typ.underlying
-        dtype = self._get_pl_dtype_for_type(underlying, fqn)
+        dtype = self._get_pa_dtype_for_type(underlying, fqn)
         if typ.collection_type is not None:
-            dtype = pl.List(dtype)
+            dtype = pa.list_(dtype)
         return dtype
 
-    def _get_pl_dtype_for_type(self, underlying: type | Windowed, fqn: str) -> Union[pl.DataType, Type[pl.DataType]]:
+    def _get_pa_dtype_for_type(self, underlying: type | Windowed, fqn: str) -> pa.DataType:
         # Polars seems to allow optional for any dtype, so we ignore it when computing dtypes
         underlying = _unwrap_optional_if_needed(underlying)
         if isinstance(underlying, Windowed):
             # If the parent window feature appears in a dataframe, then it's likely in a streaming resolver
             # Use the underlying type for that windowed feature
-            return self._get_pl_dtype_for_type(underlying._kind, fqn)
+            return self._get_pa_dtype_for_type(underlying._kind, fqn)
         if issubclass(underlying, enum.Enum):
             # For enums, require all members to have the same type
             underlying = get_enum_value_type(underlying)
         if issubclass(underlying, str):
-            return pl.Utf8
+            return pa.string()
         if issubclass(underlying, bool):
-            return pl.Boolean
+            return pa.bool_()
         if issubclass(underlying, int):
-            return pl.Int64
+            return pa.int64()
         if issubclass(underlying, float):
-            return pl.Float64
+            return pa.float64()
         if issubclass(underlying, datetime):
-            return pl.Datetime(time_unit="us", time_zone="UTC")
+            return pa.timestamp("us", "UTC")
         if issubclass(underlying, date):
-            return pl.Date
-        if dataclasses.is_dataclass(underlying) or is_namedtuple(underlying, annotated=True):
-            # Polars always treats dataclasses and namedtuples as structs
+            return pa.date64()
+        if dataclasses.is_dataclass(underlying) or _is_namedtuple(underlying, annotated=True):
             annotations = get_type_hints(underlying)
-            fields: List[pl.Field] = []
+            fields: List[pa.Field] = []
             for field_name, type_annotation in annotations.items():
                 parsed_annotation = ParsedAnnotation(underlying=type_annotation)
                 if not parsed_annotation.is_scalar:
                     raise TypeError(f"{fqn} cannot be another Features class ")
-                underlying_dtype = self._get_pl_dtype_for_type(parsed_annotation.underlying, f"{fqn}.{field_name}")
+                underlying_dtype = self._get_pa_dtype_for_type(parsed_annotation.underlying, f"{fqn}.{field_name}")
                 if parsed_annotation.collection_type is None:
                     fields.append(
-                        pl.Field(
-                            name=field_name,
-                            dtype=underlying_dtype,
+                        pa.field(
+                            field_name,
+                            underlying_dtype,
                         )
                     )
                 else:
-                    fields.append(pl.Field(name=field_name, dtype=pl.List(underlying_dtype)))
-            return pl.Struct(fields=fields)
-        _log.info(
-            f"Defaulting to polars type 'object' for fqn {fqn} of type {underlying.__name__}. Set the `polars_dtype` attribute for better specificity."
+                    fields.append(pa.field(field_name, pa.list_(underlying_dtype)))
+            return pa.struct(fields)
+        if issubclass(underlying, BaseModel):
+            fields: List[pa.Field] = []
+
+            field: ModelField
+            for field in underlying.__fields__.values():
+                field_name = field.alias if field.alias is not None else field.name
+                fields.append(
+                    pyarrow.field(
+                        metadata={"chalk_field_name": field.name},
+                        name=field_name,
+                        nullable=field.allow_none,  # noqa
+                        # abusing fqn here
+                        type=self._get_pa_dtype_for_type(field.type_, f"{fqn}.{field_name}"),
+                    )
+                )
+
+            return pa.struct(fields)
+
+        raise ValueError(
+            f"Unable to determine the PyArrow type for feature {underlying.__name__}. Please set the `pyarrow_dtype` attribute."
         )
-        return pl.Object
 
 
 FEATURE_CODEC = FeatureCodec()
