@@ -4,6 +4,7 @@ from pydantic import BaseModel
 
 from chalk.integrations.named import load_integration_variable
 from chalk.streams.base import StreamSource
+from chalk.utils.duration import Duration
 from chalk.utils.string import comma_whitespace_split
 
 
@@ -11,6 +12,13 @@ class KafkaSource(StreamSource, BaseModel, frozen=True):
     bootstrap_server: Optional[Union[str, List[str]]] = None
     topic: Optional[Union[str, List[str]]] = None
     ssl_keystore_location: Optional[str] = None
+    """
+    An S3 or GCS URI that points to the keystore file that should be
+        used for brokers. You must configure the appropriate AWS or
+        GCP integration in order for Chalk to be able to access these
+        files.
+    """
+
     client_id_prefix: str = "chalk/"
     group_id_prefix: str = "chalk/"
 
@@ -21,7 +29,7 @@ class KafkaSource(StreamSource, BaseModel, frozen=True):
             Default: PLAINTEXT.
     """
 
-    sasl_mechanism: Literal["PLAIN", "GSAPI", "SCRAM-SHA-256", "SCRAM-SHA-512"] = "PLAIN"
+    sasl_mechanism: Literal["PLAIN", "GSSAPI", "SCRAM-SHA-256", "SCRAM-SHA-512", "OAUTHBEARER"] = "PLAIN"
     """
     Authentication mechanism when security_protocol
             is configured for SASL_PLAINTEXT or SASL_SSL. Valid values are:
@@ -31,16 +39,25 @@ class KafkaSource(StreamSource, BaseModel, frozen=True):
 
     sasl_username: Optional[str] = None
     """
-    username for sasl PLAIN, SCRAM-SHA-256, or SCRAM-SHA-512 authentication.
+    Username for SASL PLAIN, SCRAM-SHA-256, or SCRAM-SHA-512 authentication.
             Default: None
     """
 
     sasl_password: Optional[str] = None
     """
-    password for sasl PLAIN, SCRAM-SHA-256, or SCRAM-SHA-512 authentication.
+    Password for SASL PLAIN, SCRAM-SHA-256, or SCRAM-SHA-512 authentication.
             Default: None
     """
+
     name: Optional[str] = None
+    """
+    The name of the integration, as configured in your web dashboard portal.
+    """
+
+    late_arrival_deadline: Duration = "infinity"
+    """
+    Messages older than this deadline will not be processed.
+    """
 
     def __init__(
         self,
@@ -51,10 +68,11 @@ class KafkaSource(StreamSource, BaseModel, frozen=True):
         client_id_prefix: Optional[str] = None,
         group_id_prefix: Optional[str] = None,
         security_protocol: Optional[str] = None,
-        sasl_mechanism: Optional[Literal["PLAIN", "GSAPI", "SCRAM-SHA-256", "SCRAM-SHA-512"]] = None,
+        sasl_mechanism: Optional[Literal["PLAIN", "GSSAPI", "SCRAM-SHA-256", "SCRAM-SHA-512", "OAUTHBEARER"]] = None,
         sasl_username: Optional[str] = None,
         sasl_password: Optional[str] = None,
         name: Optional[str] = None,
+        late_arrival_deadline: Duration = "infinity",
     ):
         super(KafkaSource, self).__init__(
             bootstrap_server=bootstrap_server
@@ -80,6 +98,7 @@ class KafkaSource(StreamSource, BaseModel, frozen=True):
             sasl_username=sasl_username or load_integration_variable(name="KAFKA_SASL_USERNAME", integration_name=name),
             sasl_password=sasl_password or load_integration_variable(name="KAFKA_SASL_PASSWORD", integration_name=name),
             name=name,
+            late_arrival_deadline=late_arrival_deadline,
         )
 
     def config_to_json(self) -> Any:
