@@ -1,7 +1,7 @@
-from datetime import timedelta
-from typing import Any, Callable, Generic, List, Literal, Mapping, Optional, Set, Type, TypeVar, Union
+from __future__ import annotations
 
-import polars as pl
+from datetime import timedelta
+from typing import TYPE_CHECKING, Any, Callable, Generic, List, Literal, Mapping, Optional, Set, Type, TypeVar, Union
 
 from chalk._validation.feature_validation import FeatureValidation
 from chalk.utils.collections import ensure_tuple
@@ -9,6 +9,9 @@ from chalk.utils.duration import Duration, parse_chalk_duration, timedelta_to_du
 
 T = TypeVar("T")
 
+if TYPE_CHECKING:
+    import polars as pl
+
 
 class WindowedInstance(Generic[T]):
     def __init__(self, values: Mapping[str, T]):
@@ -122,7 +125,7 @@ class Windowed(Generic[T], metaclass=WindowedMeta):
             etl_offline_to_online=self._etl_offline_to_online,
             encoder=self._encoder,
             decoder=self._decoder,
-            polars_dtype=self._dtype,
+            pyarrow_dtype=self._dtype,
             validations=FeatureValidation(
                 min=self._min,
                 max=self._max,
