@@ -5,7 +5,7 @@ import collections.abc
 import contextlib
 import contextvars
 import logging
-from typing import Any, List, Mapping, Optional
+from typing import Any, Dict, List, Mapping, Optional
 
 _LOGGING_CONTEXT_STACK: contextvars.ContextVar[List[Mapping[str, Any]]] = contextvars.ContextVar(
     "_LOGGING_CONTEXT_STACK"
@@ -13,7 +13,7 @@ _LOGGING_CONTEXT_STACK: contextvars.ContextVar[List[Mapping[str, Any]]] = contex
 
 
 def _recursive_merge(a: Mapping[str, Any], b: Mapping[str, Any]):
-    ans: Mapping[str, Any] = {**a}
+    ans: Dict[str, Any] = {**a}
     for k, v in b.items():
         if k not in ans or not (isinstance(ans[k], collections.abc.Mapping) and isinstance(v, collections.abc.Mapping)):
             ans[k] = v
@@ -35,7 +35,7 @@ def set_logging_context(_merge: bool = True, **extra: Any):
     Sets the context-local log context to a new value.
 
     Args:
-        merge: If true, the new context will be merged with the current context. Otherwise, the new context will overwrite the existing context.
+        _merge: If true, the new context will be merged with the current context. Otherwise, the new context will overwrite the existing context.
     """
     log_context_stack = _LOGGING_CONTEXT_STACK.get([{}])
     if _merge:
@@ -87,7 +87,7 @@ def add_logging_context(_merge: bool = True, **extra: Any):
 
     Args:
         _merge: Whether to merge the new context with the existing log context.
-        kwargs: Contextual information to add to the log record
+        extra: Contextual information to add to the log record
     """
     set_logging_context(_merge, **extra)
     try:
