@@ -31,12 +31,12 @@ client2 = Client()
 
 
 @before_all(environment="my-special-env")
-def setup_env():
+def setup_env_1():
     client.init = True
 
 
 @before_all(environment="my-special-env-2")
-def setup_env():
+def setup_env_2():
     client2.init = True
 
 
