@@ -1,10 +1,13 @@
-from typing import cast
-
 import pytest
 
 from chalk.features import features
 from chalk.sql import PostgreSQLSource
-from chalk.sql.integrations.postgres import PostgreSQLSourceImpl
+from chalk.sql._internal.integrations.postgres import PostgreSQLSourceImpl
+
+try:
+    import psycopg2
+except ImportError:
+    psycopg2 = None
 
 try:
     import psycopg2
@@ -24,10 +27,10 @@ class LibraryFeatures:
 
 @pytest.mark.skipif(psycopg2 is None, reason="psycopg2 is not installed")
 def test_table_builder():
-    source: PostgreSQLSourceImpl = cast(
-        PostgreSQLSourceImpl,
+    source = (
         PostgreSQLSource()
         .with_table(name="users", features=UserFeatures)
-        .with_table(name="libraries", features=LibraryFeatures),
+        .with_table(name="libraries", features=LibraryFeatures)
     )
+    assert isinstance(source, PostgreSQLSourceImpl)
     assert len(source.ingested_tables) == 2
