@@ -1,4 +1,3 @@
-import polars
 import pytest
 from pydantic import BaseModel
 
@@ -11,6 +10,12 @@ except ImportError:
     duckdb = None
 
 
+try:
+    import polars as pl
+except ImportError:
+    pl = None
+
+
 @features
 class StoreFeatures:
     id: str
@@ -35,10 +40,11 @@ def fn(messages: DataFrame[KafkaMessage]) -> DataFrame[StoreFeatures.id, StoreFe
     """
 
 
-@pytest.mark.skipif(duckdb is None, reason="duckdb is not installed")
+@pytest.mark.skipif(duckdb is None or pl is None, reason="duckdb is not installed or polars is not installed")
 def test_runs_sql():
+    assert pl is not None
     expected = DataFrame(
-        polars.DataFrame(
+        pl.DataFrame(
             {
                 "store_features.id": ["store1", "store2"],
                 "store_features.purchases": [10, 5],
@@ -47,7 +53,7 @@ def test_runs_sql():
     )
     actual = fn(
         DataFrame(
-            polars.DataFrame(
+            pl.DataFrame(
                 {
                     "store_id": ["store1", "store2", "store1"],
                     "amount": [4, 5, 6],
